/*
  Copyright 2003 John Knottenbelt
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
*/

#include "circularbuffer.h"
#include <cstring>
#include <iostream>

CircularBuffer::CircularBuffer(int initSize, int maxSize)
  : start(0), size(initSize), maxSize(maxSize), used(0), overflow(false)
{
  buffer = new char[size];
}

CircularBuffer::~CircularBuffer()
{
  delete [] buffer;
}


void CircularBuffer::grow()
{
  if (size == maxSize)
    return;

  int size1;
  int size2;

  if (start + used <= size) {
    size1 = used;
    size2 = 0;
  }
  else {
    size1 = size - start;
    size2 = used - size1;
  }

  int newSize = size * 2;
  if (newSize > maxSize)
    newSize = maxSize;
  
  char *newData = new char[newSize];
  memcpy(newData, buffer + start, size1);
  memcpy(newData + size1, buffer, size2);

  delete [] buffer;
  buffer = newData;
  size = newSize;
  start = 0;
}

bool CircularBuffer::write(const std::string &data)
{
  return write(data.data(), data.size());
}


bool CircularBuffer::write(const char *data, int count)
{
  while (size - used < count) {
    if (size == maxSize) {
      overflow = true;
      return false;
    }
    grow();
  }

  int tail = (start + used) % size;

  // We are using a circular buffer

  if (tail + count <= size) {
    memcpy(buffer + tail, data, count);
    used += count;
  }
  else {
    // Wrap around
    int size1 = size - tail;    //        - tail + size 
    int size2 = count - size1;	//  count + tail - size 

    memcpy(buffer + tail, data, size1);
    memcpy(buffer, data + size1, size2);
    used += count;
  }

  return true;
}

void CircularBuffer::discard(int count)
{
  if (count >= used)
    count = used;

  start = (start + count) % size;
  used -= count;
}

void CircularBuffer::getSections(char *&buffer1, int &count1, 
			       char *&buffer2, int &count2)
{
  buffer1 = buffer + start;
  buffer2 = buffer;

  if (start + used > size) {
    count1 = size - start;
    count2 = used - count1;
  }
  else {
    count1 = used;
    count2 = 0;
  }
}

bool CircularBuffer::read(char *data, int count)
{
  if (count > used)
    return false;

  char *buf1, *buf2;
  int len1, len2;

  getSections(buf1, len1, buf2, len2);

  if (count <= len1) 
    memcpy(data, buf1, count);
  else {
    memcpy(data, buf1, len1);
    memcpy(data + len1, buf2, count - len1);
  }

  start = (start + count) % size;
  used -= count;

  return true;
}

bool CircularBuffer::read(std::string &data, int count)
{
  char buf[count];
  bool result = read(buf, count);
  if (result) 
    data = std::string(buf, count);
  return result;
}

bool CircularBuffer::readLine(std::string &data, char ch)
{
  char *buf1, *buf2;
  int len1, len2;

  getSections(buf1, len1, buf2, len2);
  char *p = (char *) memchr(buf1, ch, len1);
  if (p != NULL) 
    return read(data, p - buf1 + 1);

  p = (char *) memchr(buf2, ch, len2);
  if (p == NULL)
    return false;

  int readLen = len1 + (p - buf2 + 1);
  return read(data, readLen);
}
