/*
  Copyright 2003 John Knottenbelt
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
*/

#include "config.h"
#include "unixserversocket.h"
#include <iostream>
#include <cstring>
#include <cstdlib>
#include <cerrno>
#include <cstdio>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <fcntl.h>
#include <pwd.h>
#include <grp.h>

UnixServerSocket::UnixServerSocket(const std::string &filename,
				   const std::string &socketUid,
				   const std::string &socketGid,
				   const std::string &socketMode)
  : filename(filename)
{
  fd = socket(PF_UNIX, SOCK_STREAM, 0);

  if (!fd) {
    std::cerr << "Fatal error: failed to create socket for listening:\n"
	      << strerror(errno) << std::endl;
    exit(1);
  }

  struct sockaddr_un addr;
  addr.sun_family = AF_UNIX;
  strcpy(addr.sun_path, filename.c_str());

  mode_t oldMask = umask(007);
  while (bind(fd, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
    if (errno == EADDRINUSE) {
      // Possibly old socket
      if (connect(fd, (struct sockaddr *) &addr, sizeof(addr)) >= 0) {
	close(fd);
	std::cerr << "Fatal error: dvbd already running on socket " << filename << std::endl;
	exit(1);
      }
      else 
	unlink(filename.c_str());
    }
    else {
      // Some other error
      std::cerr << "Fatal error: failed to bind socket to " << filename << "\n"
	      << strerror(errno) << std::endl;
      exit(1);
    }
  }

  if (listen(fd, 5) < 0) {
    std::cerr << "Fatal error: failed to listen on socket " << filename << "\n"
	      << strerror(errno) << std::endl;
    exit(1);
  }

  umask(oldMask);

  // set the mode
  if (socketMode != "") {
    int mode = 0;
    sscanf(socketMode.c_str(), "%o", &mode);
    if (chmod(filename.c_str(), mode) != 0) {
      std::cerr << "Fatal error: failed to set mode on socket " << filename << "\n"
		<< strerror(errno) << std::endl;
      exit(1);
    }
  }

  // set the owner
  if (socketUid != "") {
    uid_t owner;
    if (!sscanf(socketUid.c_str(), "%d", &owner)) {
      struct passwd *p = getpwnam(socketUid.c_str());
      if (p == NULL) {
	std::cerr << "Fatal error: failed to lookup user " << socketUid << " in password database\n"
		  << strerror(errno) << std::endl;
	exit(1);
      }
      owner = p->pw_uid;
    }
      
    if (chown(filename.c_str(), owner, (gid_t) -1) != 0) {
      std::cerr << "Fatal error: failed to set socket ownership " << socketUid << " on " << filename << "\n"
		<< strerror(errno) << std::endl;
      exit(1);
    }
  }

  // set the group
  if (socketGid != "") {
    uid_t group;
    if (!sscanf(socketGid.c_str(), "%d", &group)) {
      struct group *p = getgrnam(socketGid.c_str());
      if (p == NULL) {
	std::cerr << "Fatal error: failed to lookup user " << socketGid << " in group database\n"
		  << strerror(errno) << std::endl;
	exit(1);
      }
      group = p->gr_gid;
    }
      
    if (chown(filename.c_str(), (uid_t) -1, group) != 0) {
      std::cerr << "Fatal error: failed to set socket group " << socketGid << " on " << filename << "\n"
		<< strerror(errno) << std::endl;
      exit(1);
    }
  }
}

void UnixServerSocket::setNonBlocking()
{
  long flags = fcntl(fd, F_GETFL, 0); 
  flags |= O_NONBLOCK;
  fcntl(fd, F_SETFL, flags);
}


UnixServerSocket::~UnixServerSocket()
{
  close(fd);
  unlink(filename.c_str());
}

void UnixServerSocket::addSelectFDs(Select &s) const
{
  s.addReadFD(fd);
}

bool UnixServerSocket::isReady(const Select &s) const
{
  return s.readyToRead(fd);
}
  
int UnixServerSocket::accept()
{
  return ::accept(fd, NULL, 0);
}
